@@ -20,15 +20,16 @@
 
 #include "config.h"
 
-#include <config.h>
-#include <gtk/gtk.h>
-#include <glib/gi18n.h>
 #include <string.h>
 
+#include <glib/gi18n-lib.h>
+#include <gtk/gtk.h>
+
 #include "imposter.h"
 #include "impress-document.h"
-#include "ev-document-thumbnails.h"
+
 #include "ev-document-misc.h"
+#include "ev-document-thumbnails.h"
 
 struct _ImpressDocumentClass
 {
@@ -289,23 +290,17 @@ impress_document_load (EvDocument  *document,
   /* FIXME: Could we actually load uris ? */
   filename = g_filename_from_uri (uri, NULL, error);
   if (!filename)
-    {
-      g_set_error (error,
-	           EV_DOCUMENT_ERROR,
-		   EV_DOCUMENT_ERROR_INVALID,
-		   _("Remote files aren't supported"));
-      return FALSE;
-    }
+    return FALSE;
 
   imp = imp_open (filename, &err);
+  g_free (filename);
 
   if (!imp)
     {
-      g_set_error (error,
-	           EV_DOCUMENT_ERROR,
-		   EV_DOCUMENT_ERROR_INVALID,
-		   _("Invalid document"));
-      g_free (filename);
+      g_set_error_literal (error,
+                           EV_DOCUMENT_ERROR,
+                           EV_DOCUMENT_ERROR_INVALID,
+                           _("Invalid document"));
       return FALSE;
     }
   impress_document->imp = imp;
@@ -315,9 +310,13 @@ impress_document_load (EvDocument  *document,
 
 static gboolean
 impress_document_save (EvDocument  *document,
-		      const char  *uri,
-		      GError     **error)
+                       const char  *uri,
+                       GError     **error)
 {
+        g_set_error_literal (error,
+                             EV_DOCUMENT_ERROR,
+                             EV_DOCUMENT_ERROR_INVALID,
+                             "Not supported");
 	return FALSE;
 }
 
@@ -334,9 +333,9 @@ impress_document_get_n_pages (EvDocument  *document)
 
 static void
 impress_document_get_page_size (EvDocument   *document,
-			     int           page,
-			     double       *width,
-			     double       *height)
+				EvPage       *page,
+				double       *width,
+				double       *height)
 {
   ImpressDocument *impress_document = IMPRESS_DOCUMENT (document);
 
@@ -383,7 +382,7 @@ impress_document_render_pixbuf (EvDocument      *document,
   g_return_val_if_fail (IMPRESS_IS_DOCUMENT (document), NULL);
   g_return_val_if_fail (impress_document->imp != NULL, NULL);
   
-  impress_document->pagenum = rc->page;
+  impress_document->pagenum = rc->page->index;
 
   g_mutex_lock (impress_document->mutex);
   impress_document->cond = g_cond_new ();
